#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Top-level package for Vulpes."""

__author__ = "Adrien Carrel"
__email__ = "a.carrel@hotmail.fr"
__version__ = "0.2.0"

__all__ = ["utils",
           "automl"]
